@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -293,7 +293,7 @@ NOEXPORT void log_raw(SERVICE_OPTIONS *opt,
             level<=opt->log_level
 #else
             (level<=opt->log_level &&
-            global_options.option.log_stderr)
+            opt->option.log_stderr)
 #endif
             )
         ui_new_log(line);
